from majava.matchers import Matcher, Mismatch, _match


class DictContains(Matcher):
    """ Matches, if a dict contains expected subset of key-value pairs
    """

    def __init__(self, expected, *, recursive=True):
        self.expected = expected
        self.recursive = recursive

    def __repr__(self):
        return f"DictContains({repr(self.expected)})"

    def _match(self, other):
        for k, val in self.expected.items():
            if k not in other:
                raise Mismatch(other, k, f"key '{k}' not found")
            actual_val = other[k]
            try:
                if self.recursive and isinstance(val, dict):
                    val = DictContains(val)
                _match(val, actual_val)
            except Mismatch as e:
                raise e.prepend(k)


class InInterval:
    # The matcher matches, if a given value (other) lies in the interval

    def __init__(self, *value):
        self.first_value, self.second_value = value

    def __eq__(self, other):
        return self.first_value <= other <= self.second_value

    def __repr__(self):
        return f"{self.first_value}, {self.second_value}"


class IsType:
    def __init__(self, *types):
        self.types = types

    def __eq__(self, other):
        return isinstance(other, self.types)

    def __repr__(self):
        return ', '.join(it.__name__ for it in self.types)



from typing import Optional


class Mismatch(Exception):
    def __init__(self, value, path, msg):
        self.value = value
        self.path = path
        self.msg = msg

    def prepend(self, path):
        self.path = f"{path}.{self.path}" if self.path else path
        return self

    def __str__(self):
        if not self.path:
            return f"{repr(self.value)} - {self.msg}"
        return f"Value {repr(self.value)} at {repr(self.path)} does not match - {self.msg}"


class Matcher:
    _mismatch = None

    def __eq__(self, other):
        try:
            self._match(other)
            self._mismatch = None
            return True
        except Mismatch as e:
            self._mismatch = e
            return False

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def _match(self, other) -> Optional[str]:
        pass


def _match(matcher, value):
    if isinstance(matcher, Matcher):
        return matcher._match(value)
    if matcher != value:
        raise Mismatch(value, "", f"{repr(value)} != {repr(matcher)}")


class And(Matcher):
    def __init__(self, *matchers):
        self.matchers = matchers

    def __and__(self, other):
        return And(*self.matchers, other)

    def __repr__(self):
        return '&'.join(repr(it) for it in self.matchers)

    def _match(self, other):
        for matcher in self.matchers:
            _match(matcher, other)


class Or(Matcher):
    def __init__(self, *matchers):
        self.matchers = matchers

    def __or__(self, other):
        return Or(*self.matchers, other)

    def __repr__(self):
        return '|'.join(repr(it) for it in self.matchers)

    def _match(self, other):
        mismatches = []
        for matcher in self.matchers:
            try:
                _match(matcher, other)
                return
            except Mismatch as e:
                mismatches.append(e)
        raise Mismatch(other, "", ", ".join(str(i) for i in mismatches))



import pytest
from majava.basic import InInterval, IsType, DictContains


def test_ininterval():
    assert 7 == InInterval(1, 7)
    assert 10 != InInterval(1, 7)


def test_istype():
    assert 2 != IsType(str)
    assert 2 == IsType(int, str)
    assert "b" == IsType(int, str)
    assert "b" != IsType(int)
    assert "b" == IsType(str)


def test_dict__plain_match():
    assert {"a": 1, "b": 2} == DictContains({"a": 1, "b": 2})
    assert {"a": 1, "b": 2} == DictContains({"a": 1})
    assert {"a": 1, "b": 2} == DictContains({"b": 2})
    assert {"a": 1, "b": 2} == DictContains({})


@pytest.mark.parametrize("actual, expected, message", [
    ({"a": 1, "b": 2}, {"a": 1, "b": 3}, "Value 2 at 'b' does not match - 2 != 3"),
    ({"a": 1}, {"b": 1}, "Value {'a': 1} at 'b' does not match - key 'b' not found")
])
def test_dict__plain_mismatch(actual, expected, message):
    with pytest.raises(AssertionError, match=f"\n *{message}$"):
        assert actual == DictContains(expected)


@pytest.mark.parametrize("actual, expected, message", [
    (
        {"a": 1, "b": {"c": {"d": 2, "e": 3}}},
        {"b": {"c": {"e": 4}}},
        "Value 3 at 'b.c.e' does not match - 3 != 4"
    ),
])
def test_dict__nested_mismatch(actual, expected, message):
    with pytest.raises(AssertionError, match=f"\n *{message}$"):
        assert actual == DictContains(expected)
